/*!
	\file Sample NURBS project.cpp
	\brief TrimmedNURBSSurface class methods definition and implementation

	Information :
		- Author	: Vincent PRAT (vprat@ifrance.com)
		- Created	: 10-07-2001 (dd-mm-yyyy)
		- Modified	: 13-07-2001 (dd-mm-yyyy)

	Copyright  2001 Vincent PRAT

	If you make any use of this tutorial or source code, please inform me
	so as for me to know if this example is useful.

	Feedback would also be highly appreciated.
*/

#ifndef WIN32_LEAN_AND_MEAN
	#define WIN32_LEAN_AND_MEAN
#endif

/*
 * Include files
 */
#ifdef WIN32
#include <windows.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>
#include "../headers/NURBSSurface.h"

/*
 *	Global variables 
 */

// Display help ?
bool bDisplayInfo = 0;

// OpenGL Lighting stuff
float LightAmbient[4]  = {0.5f, 0.5f, 0.5f, 1.0f};
float LightDiffuse[4]  = {1.0f, 1.0f, 1.0f, 1.0f};
float LightPosition[4] = {0.0f, 0.0f, -50.0f, 1.0f};
bool bWireframe = 0;

// Performance measurement stuff (win 32 only, did not have time
// to search for a portable time function)
#ifdef WIN32
LARGE_INTEGER	g_lDisplayStartTime; 
LARGE_INTEGER	g_lDisplayStopTime;
LARGE_INTEGER	g_lCounterFrequency;
long			g_lFramesCalculated = 0;
float			fCurrentFps = 0;
#endif

// See Nehe tutorials about OpenGL stuff, I ripped a lot of stuff there
// Camera position/orientation
const float piover180 = 0.0174532925f; // To convert from degrees to radians
float heading;
float xpos = 0;
float zpos = 50;
GLfloat	yrot;
GLfloat walkbias = 0;
GLfloat walkbiasangle = 0;
GLfloat lookupdown = 0.0f;
GLfloat	z=0.0f;

// Surface rotation
bool bUseRotation = 0;
float fSurfaceRotation = 0;

// Our two surfaces
TrimmedNURBSSurface *g_pTrimmedNURBSSurface;
TrimmedNURBSSurface *g_pNURBSSurface;

/*
 * Functions used by GLUT (callbacks)
 */
void Initialize();
void Reshape(int, int);
void Display();
void Keyboard(unsigned char, int, int);
void ArrowKeys(int, int, int);

int main(int argc, char **argv)
{
// Setup GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInitWindowSize(300, 300);
	glutInitWindowPosition(10, 10);
	glutCreateWindow("Trimmed NURBS Surfaces");
	Initialize();
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Display);
	glutKeyboardFunc(Keyboard);
	glutIdleFunc(Display);
	glutSpecialFunc(ArrowKeys);

// Create surfaces
	g_pTrimmedNURBSSurface	= new TrimmedNURBSSurface( "test.txt" );
	g_pNURBSSurface			= new TrimmedNURBSSurface( "test2.txt" );

// Start measure of frame rate
#ifdef WIN32
	QueryPerformanceCounter( &g_lDisplayStartTime );
#endif

// Enter GLUT main loop
	glutMainLoop();

// Free surfaces objects
	delete g_pTrimmedNURBSSurface;
	delete g_pNURBSSurface;
	return 0;
}

/* 
 *  Keyboard function
 */ 
void Keyboard( unsigned char key, int x, int y )
{
	switch ( key ) {
    case 27:
		exit ( 0 );		
		break;
	case 'w':
		bWireframe = !bWireframe;
		if (bWireframe)
			glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
		else
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
		break;
	case 'r':
		bUseRotation = !bUseRotation;
		break;
	case 'i':
		bDisplayInfo = !bDisplayInfo;
		break;
    default:
		break;
	}
}

/*
 *  Function to handle arrow keys
 */

void ArrowKeys( int a_keys, int x, int y )
{
	switch ( a_keys ) {
    case GLUT_KEY_UP:
		xpos -= (float)sin(heading*piover180) * 0.8f;
		zpos -= (float)cos(heading*piover180) * 0.8f;
		break;
	case GLUT_KEY_DOWN:
		xpos += (float)sin(heading*piover180) * 0.8f;
		zpos += (float)cos(heading*piover180) * 0.8f;
		break;
	case GLUT_KEY_LEFT:
		heading += 2.0f;	
		yrot = heading;
		break;
	case GLUT_KEY_RIGHT:
		heading -= 2.0f;	
		yrot = heading;
		break;
	case GLUT_KEY_PAGE_UP:
		lookupdown-= 2.0f;
		break;
	case GLUT_KEY_PAGE_DOWN:
		lookupdown+= 2.0f;
		break;
	case GLUT_KEY_HOME:
		walkbias -= 4;
		break;
	case GLUT_KEY_END:
		walkbias += 4;
		break;
    default:
		break;
	}
}

void setOrthographicProjection() 
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 300, 0, 300);
	glScalef(1, -1, 1);
	glTranslatef(0, -300, 0);
	glMatrixMode(GL_MODELVIEW);
}

void resetPerspectiveProjection() 
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void renderBitmapString(
		float x, 
		float y, 
		void *font, 
		char *string) 
{  
  char *c;
  glRasterPos2f(x, y);
  for (c=string; *c != '\0'; c++) 
  {
    glutBitmapCharacter(font, *c);
  }
}


void DisplayInformationText( float fps )
{
	char pChar[100];
#ifdef WIN32
	sprintf( pChar, "Press i for information | FPS : %f", fps );
#else
	sprintf( pChar, "Press i for information" );
#endif
	glDisable( GL_LIGHTING );

	setOrthographicProjection();
	glPushMatrix();
	glLoadIdentity();

	renderBitmapString(5, 15, GLUT_BITMAP_HELVETICA_12, pChar);
	
	if (bDisplayInfo)
	{
		renderBitmapString(5, 30, GLUT_BITMAP_HELVETICA_12, "KEYS :");
		renderBitmapString(5, 45, GLUT_BITMAP_HELVETICA_12, "up/down/left/right/PgUp/PgDn : move");
		renderBitmapString(5, 60, GLUT_BITMAP_HELVETICA_12, "w : wireframe mode");
		renderBitmapString(5, 75, GLUT_BITMAP_HELVETICA_12, "i : display information");
		renderBitmapString(5, 90, GLUT_BITMAP_HELVETICA_12, "r : rotation on/off");
	}

	glPopMatrix();
	resetPerspectiveProjection();
	glEnable( GL_LIGHTING );
}



/*
 * Setup the OpenGL environment.
 */
void Initialize() {
	glClearColor(0.0, 0.0, 0.0, 0.0);					// Black Background
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
	glEnable(GL_LINE_SMOOTH);							// Line anti-aliasing
	glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations

	glShadeModel(GL_SMOOTH);
	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, 1 );
	glEnable( GL_LIGHTING );
	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE , LightDiffuse);
	glLightfv(GL_LIGHT1, GL_POSITION , LightPosition);
	glEnable(GL_LIGHT1);

	glLineWidth(0.2);
	glPointSize(4.0);
}

/*
 * Called when window is resized. Makes sure the viewport fills the window,
 */
void Reshape ( int w, int h )
{
  glViewport     ( 0, 0, w, h );
  glMatrixMode   ( GL_PROJECTION );
  glLoadIdentity ( );              
  if ( h==0 )  
     gluPerspective( 80, ( float ) w, 1.0, 5000.0 );
  else
     gluPerspective( 80, ( float ) w / ( float ) h, 1.0, 5000.0 );
  glMatrixMode   ( GL_MODELVIEW );
  glLoadIdentity ( );
}

/*
 * Main display routine.
 */
void Display() 
{
	// Clear back buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Transform the world according to camera location
	glLoadIdentity();

	GLfloat xtrans = -xpos;
	GLfloat ztrans = -zpos;
	GLfloat ytrans = -walkbias;
	GLfloat sceneroty = 360.0f - yrot;

	glRotatef(lookupdown,1.0f,0,0);
	glRotatef(sceneroty,0,1.0f,0);
	
	glTranslatef(xtrans, ytrans, ztrans);
	
	// Rotate if necessary
	if (bUseRotation)
		fSurfaceRotation+=0.3;

	// Transform and render first surface
	g_pTrimmedNURBSSurface->Begin();

	g_pTrimmedNURBSSurface->Translate( -30, 0, -40 );
	g_pTrimmedNURBSSurface->Rotate( fSurfaceRotation, 3*fSurfaceRotation, 2*fSurfaceRotation );
	
	g_pTrimmedNURBSSurface->Render();
	g_pTrimmedNURBSSurface->End();

	// Transform and render second surface
	g_pNURBSSurface->Begin();

	g_pNURBSSurface->Translate( 30, 0, -40 );
	g_pNURBSSurface->Rotate( 1.5*fSurfaceRotation, fSurfaceRotation, 3*fSurfaceRotation );
	
	g_pNURBSSurface->Render();
	g_pNURBSSurface->RenderLOD(-xtrans, -ytrans, -ztrans);
	g_pNURBSSurface->End();

#ifdef WIN32
	// One more frame has been calculated
	g_lFramesCalculated++;

	// calculate frame rate every 10 frames
	if (g_lFramesCalculated > 10)
	{
		// If we have exceeded 10 frames
		// get current counter value and frequency
		QueryPerformanceCounter( &g_lDisplayStopTime );
		QueryPerformanceFrequency( &g_lCounterFrequency );

		// calculate FPS value ( # of frames / time to calculate them )
		fCurrentFps = g_lFramesCalculated / ( (double)(g_lDisplayStopTime.QuadPart - g_lDisplayStartTime.QuadPart)/(double)g_lCounterFrequency.QuadPart );		

		// Reset counter (get current value) and reset # of frames calculated so far
		g_lFramesCalculated = 0;
		QueryPerformanceCounter( &g_lDisplayStartTime );
	}

	DisplayInformationText( fCurrentFps );
#else
	DisplayInformationText( 0 );
#endif

	// Swap buffers
	glutSwapBuffers();
}